#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <GL/glut.h>

void display(void);
void draw_cube(void);
unsigned int load_texture(const char *fname);
unsigned char *load_ppm(const char *fname, int *xsz, int *ysz);
void reshape(int x, int y);

/* texture width, height, and pixel array pointer */
int width, height;
unsigned char *pixels;

unsigned int tex_logo, tex_crate;

int main(int argc, char **argv)
{
	/* initialize glut, and create a window */
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(800, 600);
	glutCreateWindow("OpenGL window");

	/* register event handlers */
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);

	tex_logo = load_texture("ground.ppm");
	tex_crate = load_texture("crate.ppm");
	if(!tex_logo || !tex_crate) {
		return 1;
	}

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	glutMainLoop();
	return 0;
}

/* display callback */
void display(void)
{
	float lpos[] = {8, 10, 15, 1};

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);

	/* set view matrix */
	glLoadIdentity();
	gluLookAt(2, 2, 4, 0, 0, 0, 0, 1, 0);

	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	glBindTexture(GL_TEXTURE_2D, tex_crate);
	draw_cube();

	glBindTexture(GL_TEXTURE_2D, tex_logo);
	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glTexCoord2f(0, 4);
	glVertex3f(-4, -1, -4);
	glTexCoord2f(4, 4);
	glVertex3f(4, -1, -4);
	glTexCoord2f(4, 0);
	glVertex3f(4, -1, 4);
	glTexCoord2f(0, 0);
	glVertex3f(-4, -1, 4);
	glEnd();

	glutSwapBuffers();
}

void draw_cube(void)
{
	glBegin(GL_QUADS);
	/* far face (-Z) */
	glNormal3f(0, 0, -1);
	glTexCoord2f(1, 1);	glVertex3f(1, -1, -1);
	glTexCoord2f(0, 1);	glVertex3f(-1, -1, -1);
	glTexCoord2f(0, 0);	glVertex3f(-1, 1, -1);
	glTexCoord2f(1, 0);	glVertex3f(1, 1, -1);
	/* top face (+Y) */
	glNormal3f(0, 1, 0);
	glTexCoord2f(0, 1); glVertex3f(-1, 1, 1);
	glTexCoord2f(1, 1); glVertex3f(1, 1, 1);
	glTexCoord2f(1, 0); glVertex3f(1, 1, -1);
	glTexCoord2f(0, 0); glVertex3f(-1, 1, -1);
	/* bottom face (-Y) */
	glNormal3f(0, -1, 0);
	glTexCoord2f(0, 1); glVertex3f(-1, -1, -1);
	glTexCoord2f(1, 1); glVertex3f(1, -1, -1);
	glTexCoord2f(1, 0); glVertex3f(1, -1, 1);
	glTexCoord2f(0, 0); glVertex3f(-1, -1, 1);
	/* right face (+X) */
	glNormal3f(1, 0, 0);
	glTexCoord2f(0, 1); glVertex3f(1, -1, 1);
	glTexCoord2f(1, 1); glVertex3f(1, -1, -1);
	glTexCoord2f(1, 0); glVertex3f(1, 1, -1);
	glTexCoord2f(0, 0); glVertex3f(1, 1, 1);
	/* left face (-X) */
	glNormal3f(-1, 0, 0);
	glTexCoord2f(0, 1); glVertex3f(-1, -1, -1);
	glTexCoord2f(1, 1); glVertex3f(-1, -1, 1);
	glTexCoord2f(1, 0); glVertex3f(-1, 1, 1);
	glTexCoord2f(0, 0); glVertex3f(-1, 1, -1);
	/* near face (+Z) */
	glNormal3f(0, 0, 1);
	glTexCoord2f(0, 1); glVertex3f(-1, -1, 1);
	glTexCoord2f(1, 1); glVertex3f(1, -1, 1);
	glTexCoord2f(1, 0); glVertex3f(1, 1, 1);
	glTexCoord2f(0, 0); glVertex3f(-1, 1, 1);
	glEnd();
}


unsigned int load_texture(const char *fname)
{
	int width, height;
	unsigned char *pixels;
	unsigned int tex_id;

	if(!(pixels = load_ppm(fname, &width, &height))) {
		return 0;
	}
	glGenTextures(1, &tex_id);
	glBindTexture(GL_TEXTURE_2D, tex_id);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0,
			GL_RGB, GL_UNSIGNED_BYTE, pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	free(pixels); /* not needed any more */
	return tex_id;
}

unsigned char *load_ppm(const char *fname, int *xsz, int *ysz)
{
	FILE *fp;
	int i, width, height, hdrline = 0;
	unsigned char *pixels = 0;

	if(!(fp = fopen(fname, "rb"))) {
		fprintf(stderr, "failed to open pixmap: %s: %s\n", fname, strerror(errno));
		return 0;
	}

	/* read ppm header */
	while(hdrline < 3) {
		char buf[64];

		if(!fgets(buf, sizeof buf, fp))
			goto err;
		
		/* skip comments */
		if(buf[0] == '#')
			continue;

		switch(hdrline++) {
		case 0:
			/* first header line should be P6 */
			if(strcmp(buf, "P6\n") != 0)
				goto err;
			break;

		case 1:
			/* second header line contains the pixmap dimensions */
			if(sscanf(buf, "%d %d", &width, &height) != 2)
				goto err;
			break;
		}
	}

	/* allocate the image (each pixel is 3 bytes r, g, and b) */
	if(!(pixels = malloc(width * height * 3))) {
		goto err;
	}

	/* read all pixels */
	for(i=0; i<width * height * 3; i++) {
		int c = fgetc(fp);
		if(c < 0)
			goto err;
		pixels[i] = c;
	}
	fclose(fp);

	*xsz = width;
	*ysz = height;
	return pixels;

err:
	fprintf(stderr, "failed to load pixmap: %s\n", fname);
	free(pixels);
	fclose(fp);
	return 0;
}

/* reshape callback (window is resized) */
void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (float)x / (float)y, 1.0, 1000.0);
}
